/**
 * 
 */
package lm.parser.dynamic.syntax;

import java.util.ArrayList;
import java.util.HashMap;

import javax.management.InvalidAttributeValueException;

import lm.parser.InputString;
import lm.parser.ParseException;

import org.apache.log4j.Logger;

/**
 * @author Michal Jaros
 *
 */
public class Syntax {

	/**
	 * Logger
	 */
	private static final Logger log = Logger.getLogger(Syntax.class);

	/**
	 * Skladnia
	 */
	private HashMap<String, Node> syntax = null;

	/**
	 * Lista symboli pomocniczych
	 */
	private ArrayList<String> symbols = null;

	/**
	 * Gramatyka
	 */
	private String grama = null;

	/**
	 * Bierzacy indeks
	 */
	private int curentIndex = 0;

	/**
	 * Konstruktor skladni
	 * @param syntax
	 * @throws SyntaxException 
	 */
	public Syntax(String grama) throws SyntaxException{
		symbols = new ArrayList<String>();
		syntax = new HashMap<String, Node>();
		this.grama = grama.trim().replaceAll("\\s+", " ");
		log.debug(grama);
		log.debug(this.grama);
		try {
			while (curentIndex < this.grama.length()) {
				try {
					extractProduction();
				} catch (StringIndexOutOfBoundsException e) {
					throw new SyntaxException("Niespodziewany koniec gramatyki !!!", e);
				}
			}
			for (int i=0; i<symbols.size(); i++){
				if (!syntax.containsKey(symbols.get(i))){
					throw new SyntaxException("Uyto niezdefiniowanego symbolu pomocniczego '" + symbols.get(i) + "' !!!");
				}
			}
		} catch (SyntaxException e) {
			log.error("Bd skadni !!!", e);
			throw e;
		}
		symbols = null;
	}

	/**
	 * Sprawdzenie poprawnosci zdania z gramatyka
	 * @param head
	 * @param inputStr
	 * @return
	 */
	public boolean parse(String head, String inputStr){
		InputString input = new InputString(inputStr);
		try {
			Node curentNode = syntax.get(head);
			if (curentNode == null){
				throw new SyntaxException("Nie ma takiej produkcji !!!");
			}
			boolean error;
			do {
				error = !curentNode.parse(input, syntax);
				if (error){
					curentNode = curentNode.getAlternativeNode();
				}else{
					curentNode = curentNode.getNextNode();
				}
			}while(curentNode != null);
			if (input.getCurentPosition() < input.length()) {
				error = true;
			}
			if (error){
				log.error("Zdanie niepoprawne !!!");
			} else{
				log.debug("Sprawdzanie zakonczone poprawnie");
			}
			return !error;
		} catch (ParseException e) {
			log.error("Wyrarzenie niepoprawne !!!", e);
			return false;
		} catch (SyntaxException e) {
			log.error("Bd przetwarzania !!!", e);
			return false;
		} catch (RecursionException e) {
			log.error("Wyrarzenie niepoprawne !!!", e);
			return false;
		}
	}

	/**
	 * Wydobycie pojedynczej produkcji z gramatyki
	 * @throws SyntaxException - blad w definicji skladni gramatyki
	 */
	private void extractProduction() throws SyntaxException{
		String token;
		Node node;
		// nazwa produkcji
		token = extractIdentifier();
		log.debug(token);
		// '='
		if (grama.charAt(curentIndex) == '=') {
			curentIndex ++;
		} else {
			throw new SyntaxException("Oczekiwano znaku '=' !!!");
		}
		log.debug("=");
		// produkcja '.'
		node = extractExpression().getEntryNode();
		if (grama.charAt(curentIndex) == '.') {
			curentIndex ++;
		} else {
			throw new SyntaxException("Oczekiwano znaku '.' !!!");
		}
		syntax.put(token, node);
		log.debug(".");
	}

	/**
	 * @throws SyntaxException
	 */
	private Tree extractExpression() throws SyntaxException {
		Tree tree = extractTerm();
		while (grama.charAt(curentIndex) == '|') {
			curentIndex++;
			log.debug("|");
			tree.addAlternative(extractTerm());
		}
		return tree;
	}

	/**
	 * @throws SyntaxException
	 */
	private Tree extractTerm() throws SyntaxException {
		boolean endTerm = false;
		Tree tree = null;
		do {
			if (tree == null) {
				tree = extractFactor();
			} else {
				tree.addNext(extractFactor());
			}
			if (grama.charAt(curentIndex) == ' ') {
				curentIndex ++;
			}
			switch (grama.charAt(curentIndex)){
				case '=':
					throw new SyntaxException("Wystpienie '=' w niedozwolonym miejscu !!!");
				case '|':
				case '.':
				case ')':
				case ']':
				case '}':
					endTerm = true;
					break;
				default:
					break;
			}
		} while (!endTerm);
		return tree;
	}

	/**
	 * @throws SyntaxException
	 */
	private Tree extractFactor() throws SyntaxException {
		Tree tree = null;
		String token;
		if (grama.charAt(curentIndex) == ' ') {
			curentIndex ++;
		}
		switch (grama.charAt(curentIndex)){
			case '"':
				curentIndex++;
				token = extractLiteral();
				try {
					tree = new Tree(new Node(Node.TERMINAL, token));
				} catch (InvalidAttributeValueException e) {
					log.debug("Bd utworzenia wza !!!", e);
				}
				log.debug("\"");
				log.debug(token);
				log.debug("\"");
				break;
			case '(':
				// grupa
				curentIndex++;
				log.debug("group");
				tree = extractExpression();
				if (grama.charAt(curentIndex) == ')') {
					curentIndex++;
				}else{
					throw new SyntaxException("Oczekiwano znaku ')' !!!");
				}
				break;
			case '[':
				// opcja
				curentIndex++;
				log.debug("opcja");
				tree = extractExpression();
				try {
					tree.addAlternative(new Node(Node.TERMINAL, ""));
				} catch (InvalidAttributeValueException e) {
					throw new SyntaxException("Problem z utworzeniem pustego wza !!!", e);
				}
				if (grama.charAt(curentIndex) == ']') {
					curentIndex++;
				}else{
					throw new SyntaxException("Oczekiwano znaku ']' !!!");
				}
				// expression
				break;
			case '{':
				// iteracja
				curentIndex++;
				log.debug("iteracja");
				tree = extractExpression();
				tree.addNext(tree);
				try {
					tree.addAlternative(new Node(Node.TERMINAL, ""));
				} catch (InvalidAttributeValueException e) {
					throw new SyntaxException("Problem z utworzeniem pustego wza !!!", e);
				}
				if (grama.charAt(curentIndex) == '}') {
					curentIndex++;
				}else{
					throw new SyntaxException("Oczekiwano znaku '}' !!!");
				}
				// expression
				break;
			default:
				token = extractIdentifier();
				if (!symbols.contains(token)) {
					symbols.add(token);
				}
				try {
					tree = new Tree(new Node(Node.NONTERMINAL, token));
				} catch (InvalidAttributeValueException e) {
					log.debug("Bd utworzenia wza !!!", e);
				}
				log.debug(token);
				break;
		}
		return tree;
	}

	/**
	 * Wydobywal literal
	 * @return literal
	 * @throws SyntaxException - blad skladniowy
	 */
	private String extractLiteral() throws SyntaxException {
		String literal = null;
		int index = grama.indexOf('"', curentIndex);
		if (index < 0){
			throw new SyntaxException("Brak zakoczenia literau !!!");
		}
		literal = grama.substring(curentIndex, index);
		if (literal.isEmpty()) {
			log.warn("Pusty litera !!!");
		}
		curentIndex = index + 1;
		return literal;
	}

	/**
	 * Wydobywa identyfikator
	 * @return identyfikator
	 * @throws SyntaxException - blad w skladni
	 */
	private String extractIdentifier() throws SyntaxException {
		String identifier = null;
		int index = curentIndex;
		boolean first = true;
		while (index < grama.length()) {
			switch (grama.charAt(index)){
				case ' ':
					index++;
					if (first) {
						break;
					}
				case '=':
				case '|':
				case '.':
				case '"':
				case '(':
				case ')':
				case '[':
				case ']':
				case '{':
				case '}':
					identifier = grama.substring(curentIndex, index).trim();
					if (identifier.isEmpty()) {
						throw new SyntaxException("Pusty identyfikator !!!");
					}
					curentIndex = index;
					return identifier;
				default:
					first = false;
					index++;
					break;
			}
		}
		throw new SyntaxException("Brak zakoczenia identyfikatora !!!");
	}
}
